' ----------------------------------------------------------------
' ZX BASIC INTERFACE LIBRARY FOR THE BIFROST* ENGINE - RELEASE 1.2/L
'
' If you use this interface library, you must load afterwards the
' BIFROST* ENGINE and a multicolor tile set. For a detailed sample
' see file "bifrostdem.bas".
'
' Original version and further information is available at
' http://www.worldofspectrum.org/infoseekid.cgi?id=0027405
'
' Interface library developed by Einar Saukas
' Additional improvements provided by Boriel and Gedlion
' ----------------------------------------------------------------

#ifndef __LIBRARY_BIFROST__
#define __LIBRARY_BIFROST__
#pragma push(case_insensitive)
#pragma case_insensitive = true

const BIFROSTSTATIC AS UBYTE = 128
const BIFROSTDISABLED AS UBYTE = 255

' ----------------------------------------------------------------
' Activate multicolor rendering with BIFROST* ENGINE
' ----------------------------------------------------------------
#define BIFROSTstart() \
    asm                \
        call 64995     \
    end asm

' ----------------------------------------------------------------
' Deactivate multicolor rendering with BIFROST* ENGINE
' ----------------------------------------------------------------
#define BIFROSTstop() \
    asm               \
        call 65012    \
    end asm

' ----------------------------------------------------------------
' Execute HALT (wait for next frame).
'
' If an interrupt occurs while certain routines (BIFROSTdrawTileH,
' BIFROSTshowTilePosH, BIFROSTshowNextTile, etc) are under
' execution, the entire screen will "glitch".
'
' Routine BIFROSThalt can be used to avoid these problems.
' Immediately after calling it, your program will have some time
' (about 20K T) to execute a few other routines without any
' interruption.
' ----------------------------------------------------------------
#define BIFROSThalt() \
    asm               \
        halt          \
    end asm

' ----------------------------------------------------------------
' Place a multicolor tile index into the tile map. Add value
' BIFROSTSTATIC for static tile, otherwise it will be animated
'
' Parameters:
'     px: tile vertical position (0-8)
'     py: tile horizontal position (0-8)
'     tile: tile index (0-255)
' ----------------------------------------------------------------
#define BIFROSTsetTile(px, py, tile)   POKE 65281+(px)*9+(py), (tile)

' ----------------------------------------------------------------
' Obtain a multicolor tile index from the tile map
'
' Parameters:
'     px: tile vertical position (0-8)
'     py: tile horizontal position (0-8)
'
' Returns:
'     Tile index currently stored in this position
' ----------------------------------------------------------------
#define BIFROSTgetTile(px, py)   PEEK(65281+(px)*9+(py))

' ----------------------------------------------------------------
' Convert multicolor tile index into the equivalent animation group
'
' Parameters:
'     tile: tile index (0-255)
'
' Returns:
'     Animation group for animated tile, otherwise the same tile index
' ----------------------------------------------------------------
function FASTCALL BIFROSTgetAnimGroup(tile as UBYTE) AS UBYTE
    asm
        cp 128          ; compare with BIFROSTSTATIC
        ret nc          ; if (static tile) return tile
        srl a
        ld b,a
        ld a,(58780)
        and a
        ld a,b
        ret z           ; if (2 frames per animation) return tile/2
        rra             ; if (4 frames per animation) return tile/4
    end asm
end function

' ----------------------------------------------------------------
' Locate memory address that stores the multicolor attribute of a
' certain screen position inside the multicolor area
'
' Parameters:
'     lin: pixel line (16-159)
'     col: char column (1-18)
'
' Returns:
'     Memory address of the multicolor attribute
' ----------------------------------------------------------------
function FASTCALL BIFROSTfindAttrH(lin as UBYTE, col as UBYTE) as UINTEGER
    asm
        ld b,0
        ld c,a          ; BC=lin
        ld h,b
        ld l,c          ; HL=lin
        add hl,hl       ; HL=lin*2
        add hl,hl       ; HL=lin*4
        add hl,bc       ; HL=lin*5
        add hl,hl       ; HL=lin*10
        add hl,hl       ; HL=lin*20
        add hl,hl       ; HL=lin*40
        add hl,bc       ; HL=lin*41
        ex de,hl        ; DE=lin*41
        pop hl          ; RET address
        ex (sp),hl      ; H=col
        ld l,h          ; L=col
        set 6,l         ; L=col+64
        ld h,$e5        ; reference 'deltas' inside BIFROST*
        ld l,(hl)       ; L=delta
        ld h,b          ; HL=delta
        add hl,de       ; HL=lin*41 + delta
        ld de,58419     ; DE=59075 - 16*41
        add hl,de       ; HL=59075 + (lin-16)*41 + delta
    end asm
end function

' ----------------------------------------------------------------
' Reconfigure BIFROST* ENGINE to read tile images from another address
'
' Parameters:
'     addr: New tile images address
' ----------------------------------------------------------------
#define BIFROSTresetTileImages(addr)   POKE UINTEGER 58801, (addr)

' ----------------------------------------------------------------
' Reconfigure BIFROST* ENGINE to animate at 2 frames per second
' ----------------------------------------------------------------
#define BIFROSTresetAnimSlow()   POKE 59035,254

' ----------------------------------------------------------------
' Reconfigure BIFROST* ENGINE to animate at 4 frames per second
' ----------------------------------------------------------------
#define BIFROSTresetAnimFast()   POKE 59035,198

' ----------------------------------------------------------------
' Reconfigure BIFROST* ENGINE to use 2 frames per animation group
' ----------------------------------------------------------------
sub FASTCALL BIFROSTresetAnim2Frames()
    asm
        halt
        xor a
        ld (58780),a
        ld hl,128
        ld (58782),hl
    end asm
end sub

' ----------------------------------------------------------------
' Reconfigure BIFROST* ENGINE to use 4 frames per animation group
' ----------------------------------------------------------------
sub FASTCALL BIFROSTresetAnim4Frames()
    asm
        halt
        ld a,15
        ld (58780),a
        ld hl,64+256*7
        ld (58782),hl
    end asm
end sub

' ----------------------------------------------------------------
' Advanced conversions
' ----------------------------------------------------------------
#define PX2LIN(px)              (((px)+1)<<4)
#define PX2ROW(px)              (((px)<<1)+1)

#define ROW2LIN(row)            (((row)+1)<<3)
#define ROW2PX_UP(row)          ((row)>>1)
#define ROW2PX_DOWN(row)        (((row)-1)>>1)

#define LIN2ROW_UP(lin)         (((lin)>>3)-1)
#define LIN2ROW_DOWN(lin)       (((lin)-1)>>3)
#define LIN2PX_UP(lin)          (((lin)>>4)-1)
#define LIN2PX_DOWN(lin)        (((lin)-1)>>4)

#define PY2COL(py)              (((py)<<1)+1)
#define COL2PY_LEFT(col)        (((col)-1)>>1)
#define COL2PY_RIGHT(col)       ((col)>>1)

' ----------------------------------------------------------------
' Instantly draw a multicolor tile at the specified screen position
'
' Parameters:
'     row: char row (1-17)
'     col: char column (1,3,5..17)
'     tile: tile index (0-255)
'
' WARNING: Incorrect use of this routine may cause glitches
' ----------------------------------------------------------------
sub FASTCALL BIFROSTdrawTileL(row as UBYTE, col as UBYTE, tile as UBYTE)
    asm
        ld d,a          ; D=row
        pop hl          ; RET address
        pop bc          ; B=col
        ld e,b          ; E=col
        ex (sp),hl      ; H=tile
        ld a,h          ; A=tile
        di
        call 58786      ; execute 'draw_tile'
        ei
    end asm
end sub

' ----------------------------------------------------------------
' Instantly show/animate the multicolor tile currently stored in
' the specified tile map position
'
' Parameters:
'     row: char row (1,3,5..17)
'     col: char column (1,3,5..17)
'
' WARNING: Incorrect use of this routine may cause glitches
' ----------------------------------------------------------------
sub FASTCALL BIFROSTshowTilePosL(row as UBYTE, col as UBYTE)
    asm
        ld d,a          ; D=row
        pop hl          ; RET address
        ex (sp),hl      ; H=col
        ld e,h          ; E=col
        di
        call 58754      ; execute 'show_tile_pos'
        ei
    end asm
end sub

' ----------------------------------------------------------------
' Instantly show/animate the next multicolor tile currently stored
' in the tile map position, according to a pre-established drawing
' order
'
' WARNING: Incorrect use of this routine may cause glitches
' ----------------------------------------------------------------
#define BIFROSTshowNextTile() \
    asm                       \
        di                    \
        call 58727            \
        ei                    \
    end asm

' ----------------------------------------------------------------
' Instantly change the attributes in a tile area (16x16 pixels) to
' the specified value (use the same INK and PAPER values to "erase"
' a tile)
'
' Parameters:
'     row: char row (1-17)
'     col: char column (1,3,5..17)
'     attr: attribute value (0-255), INK+8*PAPER+64*BRIGHT+128*FLASH
' ----------------------------------------------------------------
sub FASTCALL BIFROSTfillTileAttrL(row as UBYTE, col as UBYTE, attr as UBYTE)
    asm
        ld d,a          ; D=row
        pop hl          ; RET address
        pop bc          ; B=col
        ld e,b          ; E=col
        ex (sp),hl      ; H=attr
        ld c,h          ; C=attr
        jp 58661        ; execute 'fill_tile_attr'
    end asm
end sub

#pragma pop(case_insensitive)
#endif
